
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Consistency check on the local geometry arrays b.ipt,b.imb,b.iup,b.idn and
* the gauge field arrays b.u,b.ud of a new block b
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "block.h"
#include "global.h"

static int ix_test[VOLUME];


static int check_unity(su3 *u)
{
   int i;
   float *rv;
   su3 v;

   v=*u;

   v.c11.re-=1.0f;
   v.c22.re-=1.0f;
   v.c33.re-=1.0f;

   rv=(float*)(&v);

   for (i=0;i<18;i++)
      if (rv[i]!=0.0f)
         return 1;

   return 0;
}


static int check_unity_dble(su3_dble *u)
{
   int i;
   double *rv;
   su3_dble v;

   v=*u;

   v.c11.re-=1.0;
   v.c22.re-=1.0;
   v.c33.re-=1.0;

   rv=(double*)(&v);

   for (i=0;i<18;i++)
      if (rv[i]!=0.0)
         return 1;

   return 0;
}


int main(int argc,char *argv[])
{
   int my_rank;
   int bo[4],bs[4];
   int ix,iy,is,itest;
   int x0,x1,x2,x3,mu,x[4];
   int y0,y1,y2,y3;
   su3 *ub,*rub;
   su3_dble *udb,*rudb;
   block_t b;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);
      fin=freopen("check1.in","r",stdin);

      printf("\n");
      printf("Consistency checks on the local arrays of a new block b\n");
      printf("-------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bo %d %d %d %d\n",&bo[0],&bo[1],&bo[2],&bo[3]);
      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bo = %d %d %d %d\n",bo[0],bo[1],bo[2],bo[3]);
      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bo,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   geometry();
   alloc_u();
   alloc_ud();

   alloc_blk(&b,bo,bs,1,1,0,0);

   error((bo[0]!=b.bo[0])||(bo[1]!=b.bo[1])||
         (bo[2]!=b.bo[2])||(bo[3]!=b.bo[3])||
         (bs[0]!=b.bs[0])||(bs[1]!=b.bs[1])||
         (bs[2]!=b.bs[2])||(bs[3]!=b.bs[3]),1,"main [check1.c]",
         "The block parameters are not properly assigned");

   error((b.vol!=(bs[0]*bs[1]*bs[2]*bs[3]))||(2*b.volh!=b.vol),1,
         "main [check1.c]","b.vol or b.volh is incorrect");

   error((b.ns!=0)||(b.nsd!=0)||(b.s!=NULL)||(b.sd!=NULL),1,
         "main [check1.c]","The spinor field parameters are incorrectly set");

   for (ix=0;ix<VOLUME;ix++)
      ix_test[ix]=0;

   itest=0;

   for (ix=0;ix<b.vol;ix++)
   {
      iy=b.imb[ix];

      if ((iy<0)||(iy>=VOLUME))
         itest=1;
      else
      {
         ix_test[iy]+=1;
         if (ix_test[iy]>1)
            itest=2;
      }
   }

   error(itest==1,1,"main [check1.c]","b.imb is out of range");
   error(itest==2,1,"main [check1.c]","b.imb is not one-to-one");

   for (x0=0;x0<b.bs[0];x0++)
   {
      for (x1=0;x1<b.bs[1];x1++)
      {
         for (x2=0;x2<b.bs[2];x2++)
         {
            for (x3=0;x3<b.bs[3];x3++)
            {
               x[0]=x0;
               x[1]=x1;
               x[2]=x2;
               x[3]=x3;

               y0=bo[0]+x0;
               y1=bo[1]+x1;
               y2=bo[2]+x2;
               y3=bo[3]+x3;

               iy=ipt[y3+L3*y2+L2*L3*y1+L1*L2*L3*y0];
               ix=ipt_blk(&b,x);

               if ((ix<0)||(ix>b.vol))
                  itest=1;
               else
               {
                  if (iy!=b.imb[ix])
                     itest=2;

                  is=(x0+x1+x2+x3+bo[0]+bo[1]+bo[2]+bo[3])%2;

                  if (((is==0)&&(ix>=b.volh))||((is!=0)&&(ix<b.volh)))
                     itest=3;

                  for (mu=0;mu<4;mu++)
                  {
                     if ((x[mu]+1)<b.bs[mu])
                     {
                        if (b.imb[b.iup[ix][mu]]!=iup[iy][mu])
                           itest=4;
                     }
                     else
                     {
                        if (b.iup[ix][mu]!=b.vol)
                           itest=5;
                     }

                     if (x[mu]>0)
                     {
                        if (b.imb[b.idn[ix][mu]]!=idn[iy][mu])
                           itest=6;
                     }
                     else
                     {
                        if (b.idn[ix][mu]!=b.vol)
                           itest=7;
                     }
                  }
               }
            }
         }
      }
   }

   error(itest==1,1,"main [check1.c]",
         "b.ipt is out of range");
   error(itest==2,1,"main [check1.c]",
         "The block is not properly embedded");
   error(itest==3,1,"main [check1.c]",
         "b.ipt does not respect the even-odd ordering");
   error(itest==4,1,"main [check1.c]",
         "b.iup is incorrect");
   error(itest==5,1,"main [check1.c]",
         "b.iup is incorrect at the block boundary");
   error(itest==6,1,"main [check1.c]",
         "b.idn is incorrect");
   error(itest==7,1,"main [check1.c]",
         "b.idn is incorrect at the block boundary");

   ub=b.u[0];
   udb=b.ud[0];

   for (ix=0;ix<b.volh;ix++)
   {
      for (mu=0;mu<8;mu++)
      {
         rub=&(b.u[ix][mu]);
         rudb=&(b.ud[ix][mu]);

         if (check_unity(rub)!=0)
            itest=1;
         if ((rub-ub)!=(8*ix+mu))
            itest=2;
         if (check_unity_dble(rudb)!=0)
            itest=3;
         if ((rudb-udb)!=(8*ix+mu))
            itest=4;
      }
   }

   error(itest==1,1,"main [check1.c]",
         "*b.u is not properly initialized");
   error(itest==2,1,"main [check1.c]",
         "b.u is not aligned");
   error(itest==3,1,"main [check1.c]",
         "*b.ud is not properly initialized");
   error(itest==4,1,"main [check1.c]",
         "b.ud is not aligned");

   error_chk();

   if (my_rank==0)
   {
      printf("\n");
      printf("The block is correctly mapped by the local arrays\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
