
/*******************************************************************************
*
* File check11.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program spinor_prod_gamma[]()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "dfl.h"
#include "global.h"


static su3_vector_dble mul_cplx(complex_dble z,su3_vector_dble s)
{
   su3_vector_dble r;

   r.c1.re=z.re*s.c1.re-z.im*s.c1.im;
   r.c1.im=z.im*s.c1.re+z.re*s.c1.im;
   r.c2.re=z.re*s.c2.re-z.im*s.c2.im;
   r.c2.im=z.im*s.c2.re+z.re*s.c2.im;
   r.c3.re=z.re*s.c3.re-z.im*s.c3.im;
   r.c3.im=z.im*s.c3.re+z.re*s.c3.im;

   return r;
}


static spinor_dble mul_gamma(int mu,spinor_dble s)
{
   spinor_dble r;
   complex_dble i,m_i,m_1;

   i.re=0.0;
   i.im=1.0;

   m_i.re=0.0;
   m_i.im=-1.0;

   m_1.re=-1.0;
   m_1.im=0.0;

   if (mu==0)
   {
      r.c1=mul_cplx(m_1,s.c3);
      r.c2=mul_cplx(m_1,s.c4);
      r.c3=mul_cplx(m_1,s.c1);
      r.c4=mul_cplx(m_1,s.c2);
   }
   else if (mu==1)
   {
      r.c1=mul_cplx(m_i,s.c4);
      r.c2=mul_cplx(m_i,s.c3);
      r.c3=mul_cplx(i,s.c2);
      r.c4=mul_cplx(i,s.c1);
   }
   else if (mu==2)
   {
      r.c1=mul_cplx(m_1,s.c4);
      r.c2=s.c3;
      r.c3=s.c2;
      r.c4=mul_cplx(m_1,s.c1);
   }
   else if (mu==3)
   {
      r.c1=mul_cplx(m_i,s.c3);
      r.c2=mul_cplx(i,s.c4);
      r.c3=mul_cplx(i,s.c1);
      r.c4=mul_cplx(m_i,s.c2);
   }
   else
   {
      r.c1=s.c1;
      r.c2=s.c2;
      r.c3=mul_cplx(m_1,s.c3);
      r.c4=mul_cplx(m_1,s.c4);
   }

   return r;
}


int main(int argc,char *argv[])
{
   int my_rank,ix,mu;
   double rv,dev0[2],dev1[2];
   complex_dble sp0[2],sp1[2];
   spinor_dble **wsd;
   FILE *flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check11.log","w",stdout);
      
      printf("\n");
      printf("Check of the program spinor_prod_gamma[]()\n");
      printf("------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      fflush(flog);
   }

   start_ranlux(0,123456);   
   alloc_sd(3);

   message("Choose random fields, compute (psi,gamma_mu*chi)\n\n");
   
   wsd=reserve_wsd(3);
   rv=sqrt((double)(VOLUME));

   for (mu=0;mu<4;mu++)
   {
      random_sd(VOLUME,wsd[0],1.0);
      random_sd(VOLUME,wsd[1],1.0);
      normalize_dble(VOLUME,0,wsd[0]);
      normalize_dble(VOLUME,0,wsd[1]);
      
      spinor_prod_gamma[mu](VOLUME,wsd[0],wsd[1],sp0);

      sp1[0]=spinor_prod_dble(VOLUME,0,wsd[0],wsd[1]);

      for (ix=0;ix<VOLUME;ix++)
         wsd[2][ix]=mul_gamma(mu,wsd[1][ix]);

      sp1[1]=spinor_prod_dble(VOLUME,0,wsd[0],wsd[2]);

      dev0[0]=fabs(sp0[0].re-sp1[0].re)+fabs(sp0[0].im-sp1[0].im);
      dev0[1]=fabs(sp0[1].re-sp1[1].re)+fabs(sp0[1].im-sp1[1].im);

      MPI_Reduce(dev0,dev1,2,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

      message("mu = %d, absolute deviations: sp[0] = %.1e, sp[1] = %.1e\n",
              mu,rv*dev1[0],rv*dev1[1]);
   }
      
   error_chk();
   
   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }
   
   MPI_Finalize();   
   exit(0);
}
