
/*******************************************************************************
*
* File check5.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program dfl_subspace()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "flags.h"
#include "block.h"
#include "linalg.h"
#include "linsolv.h"
#include "sw_term.h"
#include "dirac.h"
#include "dfl.h"
#include "global.h"


static void check_basis(int Ns,double *dev0,double *dev1)
{
   int nb,isw,i,j;
   double dev,x[2],y[2];
   complex_dble z;
   block_t *b,*bm;

   b=blk_list(DFL_BLOCKS,&nb,&isw);
   bm=b+nb;

   x[0]=0.0;
   x[1]=0.0;

   for (;b<bm;b++)
   {
      for (i=1;i<=Ns;i++)
      {
         for (j=1;j<=i;j++)
         {
            z=spinor_prod_dble((*b).vol,0,(*b).sd[i],(*b).sd[j]);
            dev=sqrt(z.re*z.re+z.im*z.im);

            if (i==j)
               dev=fabs(1.0-dev);

            if (dev>x[0])
               x[0]=dev;
         }

         assign_s2sd((*b).vol,(*b).s[i],(*b).sd[0]);
         z.re=-1.0;
         z.im=0.0;
         mulc_spinor_add_dble((*b).vol,(*b).sd[0],(*b).sd[i],z);         
         dev=norm_square_dble((*b).vol,0,(*b).sd[0]);
         dev=sqrt(dev);

         if (dev>x[1])
            x[1]=dev;         
      }
   }

   if (NPROC>1)
   {
      MPI_Reduce(x,y,2,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(y,2,MPI_DOUBLE,0,MPI_COMM_WORLD);

      (*dev0)=y[0];
      (*dev1)=y[1];
   }
   else
   {
      (*dev0)=x[0];
      (*dev1)=x[1];
   }
}


int main(int argc,char *argv[])
{
   int my_rank,i,j;
   int bs[4],Ns,iwd0;
   double dev,dev0,dev1;
   complex **v;
   spinor **mds;
   spinor_dble **wsd;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check5.log","w",stdout);
      fin=freopen("check5.in","r",stdin);
      
      printf("\n");
      printf("Check of the program dfl_subspace()\n");
      printf("-----------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("Ns","%d",&Ns);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("Ns = %d\n\n",Ns);

      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
   
   start_ranlux(0,123456);   
   geometry();
   alloc_s(Ns);
   alloc_sd(1);

   v=amalloc(Ns*sizeof(*v),3);
   error(v==NULL,1,"main [check10.c]","Unable to allocate auxiliary array");
   
   set_lat_parms(0.0,0.12,1.0);
   set_dfl_parms(bs,Ns,4,1,1,1,1,1,1,0.12,1.0e-3,1.0e-6,1.0e-12,0,0);
   
   mds=reserve_ws(Ns);
   wsd=reserve_wsd(1);
   iwd0=(wsd[0]-psd[0][0])/NSPIN;

   for (i=0;i<Ns;i++)
      random_s(VOLUME,mds[i],1.0f);

   dfl_subspace(mds);
   check_basis(Ns,&dev0,&dev1);

   message("Orthonormality of the basis vectors: %.1e\n",dev0);
   message("Single-precision basis vectors:      %.1e\n",dev1);

   dev0=0.0;

   for (i=0;i<Ns;i++)
   {
      assign_s2sd(VOLUME,mds[i],wsd[0]);
      dfl_Qproj_dble(iwd0,iwd0);

      dev=norm_square_dble(VOLUME,1,wsd[0])/
         (double)(norm_square(VOLUME,1,mds[i]));
      if (dev>dev0)
         dev0=dev;
   }

   message("Consistency with the input vectors:  %.1e\n\n",sqrt(dev0));

   dfl_global_modes(v);
   dev1=0.0;

   for (i=0;i<Ns;i++)
   {
      dev=(double)(norm_square(VOLUME,1,mds[i]));
      j=(mds[i]-ps[0][0])/NSPIN;      
      dfl_sub_v2s(v[i],j);
      dev=(double)(norm_square(VOLUME,1,mds[i]))/dev;

      if (dev>dev1)
         dev1=dev;
   }
   
   message("Check of dfl_global_modes:\n");
   message("Comparison with input fields:   %.1e\n\n",sqrt(dev1));   
   
   error_chk();   
   if (my_rank==0)
      fclose(flog);
   
   MPI_Finalize();   
   exit(0);
}
