
/*******************************************************************************
*
* File check17.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Selection of the active links
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "block.h"
#include "dirac.h"
#include "hmc.h"
#include "global.h"


static int is_on_bnd(block_t *b,int ix)
{
   int mu;

   if (ix>=(*b).vol)
      return 1;
   
   for (mu=0;mu<4;mu++)
      if (((*b).iup[ix][mu]>=(*b).vol)||((*b).idn[ix][mu]>=(*b).vol))
         return 1;

   return 0;
}


int main(int argc,char *argv[])
{
   int my_rank,bs[4],n,nb,isw;
   int *lk,*lkm,*nl,nlks,vol,ierr;
   int ix,iy,mu,is,ibx,iby;
   block_t *b,*bm;
   hmc_t *hflds;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check17.log","w",stdout);
      fin=freopen("check17.in","r",stdin);
      
      printf("\n");
      printf("Selection of the active links\n");
      printf("-----------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);      
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   
   start_ranlux(0,12346);
   geometry();
   alloc_ud();
   alloc_bgr(HMC_BLOCKS,bs);

   nlks=
      (bs[0]-1)*(bs[1]-2)*(bs[2]-2)*(bs[3]-2)+
      (bs[0]-2)*(bs[1]-1)*(bs[2]-2)*(bs[3]-2)+
      (bs[0]-2)*(bs[1]-2)*(bs[2]-1)*(bs[3]-2)+
      (bs[0]-2)*(bs[1]-2)*(bs[2]-2)*(bs[3]-1);

   vol=bs[0]*bs[1]*bs[2]*bs[3];
   
   nl=amalloc(4*vol*sizeof(int),3);
   error(nl==NULL,1,"main [check17.c]","Unable to allocate auxiliary array");
   
   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;
   hflds=hmcflds();
   ierr=0;

   for (;b<bm;b++)
   {
      if (((*hflds).nlks!=nlks)||((*b).vol!=vol))
         ierr=1;
      else
      {
         for (n=0;n<(4*vol);n++)
            nl[n]=0;

         lk=(*hflds).lks;
         lkm=lk+(*hflds).nlks;

         for (;lk<lkm;lk++)
         {
            n=(*lk);
            
            if ((n<0)||(n>=(4*vol)))
               ierr=2;
            else
            {
               nl[n]+=1;
               ix=(n/8)+(*b).volh;
               mu=n%8;
               is=mu%2;
               mu/=2;

               if (is==0)
                  iy=(*b).iup[ix][mu];
               else
                  iy=(*b).idn[ix][mu];

               ibx=is_on_bnd(b,ix);
               iby=is_on_bnd(b,iy);

               if ((ibx==1)&&(iby==1))
                  ierr=3;
            }
         }

         for (n=0;n<(4*vol);n++)
            if (nl[n]>1)
               ierr=4;
      }
      
      hflds+=1;
   }

   MPI_Reduce(&ierr,&n,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   
   if (my_rank==0)
   {
      if (n==0)
         printf("No errors detected\n\n");
      else if (n==1)
         printf("Error: number of points or active links are incorrect\n\n");
      else if (n==2)
         printf("Error: active link label is out of range\n\n");
      else if (n==3)
         printf("Error: some active links stick out\n\n");
      else if (n==4)
         printf("Error: active link label is not one-to-one\n\n");
      else
         printf("Error: unexpected error number\n\n");
      fclose(flog);
   }

   MPI_Finalize();    
   exit(0);
}
