
/*******************************************************************************
*
* File check8.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of assign_chi2sd, assign_sd2chi and sub_assign_PdOs
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "block.h"
#include "linalg.h"
#include "dirac.h"
#include "sw_term.h"
#include "hmc.h"
#include "global.h"
   

int main(int argc,char *argv[])
{
   int my_rank,nb,nbh,isw;
   int bs[4],n,i,j;
   double d,dmax,dmax_all,*ra,*rb;
   complex_dble z;
   block_t *b;
   bndry_t *bb;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check8.log","w",stdout);
      fin=freopen("check8.in","r",stdin);
      
      printf("\n");
      printf("Check of assign_chi2sd, assign_sd2chi and sub_assign_PdOs\n");
      printf("---------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);      
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);       
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_sd(4);
   alloc_bgr(HMC_BLOCKS,bs);
   b=blk_list(HMC_BLOCKS,&nb,&isw);
   nbh=nb/2;
   
   for (n=0;n<nb;n++)
   {
      (*b).shf&=(~0x20);
      alloc_bnd(b,0,2);
      (*b).shf|=0x20;
      b+=1;
   }
   
   random_ud();
   assign_ud2udbgr(HMC_BLOCKS);

   random_chi();
   assign_chi2sd(0);
   assign_chi2sd(1);
   assign_sd2chi(0);
   random_sd(VOLUME,psd[0][0],1.0);
   assign_chi2sd(0);

   z.re=-1.0;
   z.im=0.0;
   mulc_spinor_add_dble(VOLUME,psd[0][0],psd[1][0],z);
   dmax=
      norm_square_dble(VOLUME,1,psd[0][0])/
      norm_square_dble(VOLUME,1,psd[1][0]);

   if (my_rank==0)
   {
      printf("Consistency of assign_chi2sd and assign_sd2chi:\n");
      printf("Maximal relative deviation = %.1e\n\n",sqrt(dmax));
   }

   random_sd(VOLUME,psd[0][0],1.0);
   assign_sd2chi(0);
   assign_chi2sd(1);
   b=blk_list(HMC_BLOCKS,&nb,&isw);
   b+=isw*nbh;
   dmax=0.0;
   
   for (n=0;n<nbh;n++)
   {
      assign_sd2sdb(b,ALL_PTS,0,0);
      assign_sd2sdb(b,ALL_PTS,1,1);
      Qbnds_dble(b,0,0);
      Qbnds_dble(b,1,1);

      for (i=0;i<8;i++)
      {
         bb=(*b).bb+i;

         ra=(double*)((*bb).sd[0]);
         rb=(double*)((*bb).sd[1]);

         for (j=0;j<(12*(*bb).vol);j++)
         {
            d=fabs(ra[j]-rb[j]);
            if (d>dmax)
               dmax=d;
         }
      }

      b+=1;
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);   

   if (my_rank==0)
   {
      printf("Comparison of DdOs*PdOs*psi with DdOs*psi:\n");
      printf("Maximal absolute deviation = %.1e\n\n",dmax_all);
   }

   random_sd(VOLUME,psd[2][0],1.0);
   assign_sd2sd(VOLUME,psd[2][0],psd[3][0]);
   z.re=1.0;
   z.im=0.0;
   mulc_spinor_add_dble(VOLUME,psd[2][0],psd[1][0],z);
   sub_assign_PdOs(2,0);
   z.re=-1.0;
   mulc_spinor_add_dble(VOLUME,psd[2][0],psd[3][0],z);
   dmax=
      norm_square_dble(VOLUME,1,psd[2][0])/
      norm_square_dble(VOLUME,1,psd[3][0]);

   if (my_rank==0)
   {
      printf("Comparison of assign_chi2sd, assign_sd2chi and ");
      printf("sub_assign_PdOs:\n");
      printf("Maximal relative deviation = %.1e\n\n",sqrt(dmax));
   }   
   
   random_chi();
   assign_chi2sd(0);
   dmax=norm_square_dble(VOLUME,1,psd[0][0]);
   sub_assign_PdOs(0,0);
   dmax=norm_square_dble(VOLUME,1,psd[0][0])/dmax;

   if (my_rank==0)
      printf("||chi-PdOs*chi||/||chi|| = %.1e\n\n",sqrt(dmax));

   random_sd(VOLUME,psd[0][0],1.0);
   set_sd2zero(VOLUME,psd[1][0]);
   set_sd2zero(VOLUME,psd[2][0]);
   sub_assign_PdOs(1,0);
   sub_assign_PdOs(2,1);
   z.re=1.0;
   z.im=0.0;
   mulc_spinor_add_dble(VOLUME,psd[2][0],psd[1][0],z);
   dmax=
      norm_square_dble(VOLUME,1,psd[2][0])/
      norm_square_dble(VOLUME,1,psd[1][0]);

   error_chk();
 
   if (my_rank==0)
   {
      printf("||PdOs*psi-PdOs*PdOs*psi||/||PdOs*psi|| = %.1e\n\n",sqrt(dmax));
      fclose(flog);   
   }      
   
   MPI_Finalize();   
   exit(0);
}
