
/*******************************************************************************
*
* File time2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of hmc_frg()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "hmc.h"
#include "global.h"


int main(int argc,char *argv[])
{
   int my_rank,n,count;
   int bs[4],nb,isw,ib;
   double wt1,wt2,wdt;
   block_t *b;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   
   if (my_rank==0)
   {
      flog=freopen("time2.log","w",stdout);
      fin=freopen("time2.in","r",stdin);
      
      printf("\n");
      printf("Timing of hmc_frg()\n");
      printf("-------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);      
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,12345);
   geometry();
   alloc_ud();
   alloc_bgr(HMC_BLOCKS,bs);

   random_ud();
   assign_ud2udbgr(HMC_BLOCKS);
   b=blk_list(HMC_BLOCKS,&nb,&isw);

   n=500000/VOLUME;
   if (n<2)
      n=2;
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();     
   for (count=0;count<n;count++)
   {
      for (ib=0;ib<nb;ib++)
         hmc_frg(ib);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();

   wdt=1.0e6*(wt2-wt1)/(double)(n*VOLUME);
   error_chk();
   
   if (my_rank==0)
   {
      printf("Time per lattice point: %4.3f micro sec\n\n",wdt);
      fclose(flog);
   }   

   MPI_Finalize();   
   exit(0);
}
