
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Exporting and importing configurations
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "misc.h"
#include "start.h"
#include "update.h"
#include "global.h"

static su3_dble u[4*VOLUME];


int main(int argc,char *argv[])
{
   int my_rank,nsize,i;
   float beta;
   double *ru,*rv,d,dmax;
   su3_dble *v;
   char cnfg_dir[NAME_SIZE],name[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      fin=freopen("check3.in","r",stdin);
      
      printf("\n");
      printf("Exporting and importing configurations\n");
      printf("--------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("cnfg_dir","%s\n",cnfg_dir);
      fclose(fin);
      
      fflush(flog);
   }

   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   
   beta=5.7f;
   start_ranlux(0,123456);
   geometry();
   alloc_u();
   alloc_ud();
   random_u();

   update(beta,4,0,8);
   assign_u2ud();

   check_dir_root(cnfg_dir);   
   nsize=name_size("%s/testcnfg",cnfg_dir);
   error_root(nsize>=NAME_SIZE,1,"main [check3.c]","cnfg_dir name is too long");
   sprintf(name,"%s/testcnfg",cnfg_dir);
   export_cnfg(name);

   v=pud[VOLUME/2][0];

   for (i=0;i<(4*VOLUME);i++)
      u[i]=v[i];

   update(beta,4,0,8);
   assign_u2ud();

   import_cnfg(name);

   dmax=0.0;
   ru=(double*)(u);
   rv=(double*)(v);

   for (i=0;i<(72*VOLUME);i++)
   {
      d=fabs((*ru)-(*rv));
      if (d>dmax)
         dmax=d;
      ru+=1;
      rv+=1;
   }

   error_chk();
   MPI_Reduce(&dmax,&d,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("Exported gauge configuration to file %s\n",name);
      printf("Then reimported configuration from there\n\n");
      printf("Maximal deviation = %.1e ",d);
      printf("(should be exactly equal to 0.0)\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
