
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Test of the Chebyshev approximation of ln(1+x)/x 
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "random.h"

#define N 100000

static double f(double x)
{
   double y;
   
   if (fabs(x)<=1.0e-3)
   {
      y=1.0/5.0;
      y=x*y-1.0/4.0;
      y=x*y+1.0/3.0;
      y=x*y-1.0/2.0;
      y=x*y+1.0;
   }
   else
      y=log(1.0+x)/x;

   return(y);
}


int main(void)
{
   int i;
   float c[9],r[N],u,v,w,x,y,z;
   double d,dmax;

   printf("\n");
   printf("Check of the Chebyshev approximation of ln(1+x)/x\n");
   printf("-------------------------------------------------\n");
   printf("\n");

   c[0]= 0.961203933f;
   c[1]=-0.167358673f;
   c[2]= 0.019275252f;
   c[3]=-0.002490170f;
   c[4]= 0.000342671f;
   c[5]=-0.000049081f;
   c[6]= 0.000007227f;
   c[7]=-0.000001086f;
   c[8]= 0.000000166f;

   ranlxs(r,N);

   dmax=0.0;
   
   for (i=0;i<N;i++)
   {
      x=-0.25f+0.75f*r[i];

      z=(4.0f*x-0.5f)/0.75f;
      u=c[8];
      v=c[7];
      w=z*u+v;
      v=u-c[6];
      u=z*w-v;
      v=w-c[5];
      w=z*u-v;
      v=u-c[4];
      u=z*w-v;
      v=w-c[3];
      w=z*u-v;
      v=u-c[2];
      u=z*w-v;
      v=w-c[1];
      w=z*u-v;
      v=u-c[0];
      y=0.5f*z*w-v;

      d=f((double)(x))/(double)(y)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
   }
   
   printf("Range -0.25<=x<=0.5, %d uniformly distributed values\n",N);
   printf("Maximal relative error of the fit: %4.2e\n\n",dmax);   
   exit(0);
}

