
/*******************************************************************************
*
* File check7.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of det_pauli_dble
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "random.h"
#include "start.h"
#include "sw_term.h"
#include "eva.h"

#define NM 10000


static double slow_det(pauli_dble *m)
{
   int i,j,k;
   double d[6];
   complex_dble a[36],v[36];

   k=6;
   
   for (i=0;i<6;i++)
   {
      a[6*i+i].re=(*m).u[i];

      for (j=i+1;j<6;j++)
      {
         a[6*i+j].re= (*m).u[k];
         a[6*j+i].re= (*m).u[k];
         k+=1;
         a[6*i+j].im= (*m).u[k];
         a[6*j+i].im=-(*m).u[k];
         k+=1;
      }
   }

   jacobi2_dble(6,a,d,v);

   return d[0]*d[1]*d[2]*d[3]*d[4]*d[5];
}


int main(void)
{
   int n,k,is,itot;
   double fact,d1,d2,d3,d1max,d2max;
   pauli_dble *md,*imd;

   printf("\n");
   printf("Check of det_pauli_dble\n");
   printf("-----------------------\n\n");

   md=amalloc(2*NM*sizeof(pauli_dble),4);
   error(md==NULL,1,
         "main [check7.c]","Unable to allocate auxiliary arrays");
   
   rlxd_init(1,1234);
   imd=md+NM;
   itot=0;
   d1max=0.0;
   d2max=0.0;
   fact=sqrt(2.0);

   for (n=0;n<NM;n++)
   {
      gauss_dble((*md).u,36);

      for (k=0;k<6;k++)
         (*md).u[k]*=fact;
         
      is=inv_pauli_dble(md,imd);
      
      if (is==0)
      {
         d1=det_pauli_dble(md);
         d2=slow_det(md);
         d3=det_pauli_dble(imd)*d1;

         d1=fabs(1.0-d1/d2);
         d2=fabs(1.0-d3);

         if (d1>d1max)
            d1max=d1;
         if (d2>d2max)
            d2max=d2;
      }
      else
         itot+=1;
      
      md+=1;
      imd+=1;
   }

   printf("%d gaussian random matrices, %d inversion failures\n",NM,itot);
   printf("det(M):          maximal deviation = %.1e\n",d1max);
   printf("det(M)*det(1/M): maximal deviation = %.1e ",d2max);
   printf("(safe cases only)\n\n");
   exit(0);
}
