
/*******************************************************************************
*
* File time1.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of mul_pauli
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "random.h"
#include "sw_term.h"
#include "start.h"

#define N 20000000

int main(void)
{
   int count;
   double t1,t2,dt;
   spinor *sb,*s;
   pauli *mb,*m;

   printf("\n");
   printf("Timing of mul_pauli\n");
   printf("-------------------\n\n");

   sb=amalloc(3*sizeof(spinor),7);
   mb=amalloc(2*sizeof(pauli),7);
   error((sb==NULL)||(mb==NULL),1,"main [time1.c]",
         "Unable to allocate auxiliary arrays");

   rlxs_init(0,23456);

   for (s=sb;s<(sb+3);s++)
      ranlxs((float*)(s),24);

   for (m=mb;m<(mb+2);m++)
      ranlxs((*m).u,36);

   t1=(double)clock();
   for (count=0;count<N;count++)
   {
      mul_pauli(mb,(weyl*)(sb),(weyl*)(sb+1));
      mul_pauli(mb+1,(weyl*)(sb)+1,(weyl*)(sb+1)+1);
   }
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(2*N);

#if (defined SSE3)
   printf("Using SSE3 instructions:\n");
#elif (defined SSE2)
   printf("Using SSE2 instructions:\n");
#elif (defined SSE)
   printf("Using SSE instructions:\n");
#endif

   printf("The time per multiplication is %4.3f micro sec",dt);
   printf(" (%d Mflops [%d bit arithmetic])\n\n",
          (int)(276.0/dt),(int)(sizeof(spinor)/3));

   exit(0);
}
