
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Normalization of the link variables after multiple updates
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "update.h"
#include "global.h"

#define RENORMALIZE 12


static void dev(su3 *u,float *d1,float *d2)
{
   int i;
   float d;
   complex det1,det2,det3,det;
   su3 v,w;

   _su3_dagger(v,(*u));
   _su3_times_su3(w,v,(*u));

   w.c11.re-=1.0f;
   w.c22.re-=1.0f;
   w.c33.re-=1.0f;

   *d1=0.0f;

   for (i=0;i<18;i++)
   {
      d=*((float*)(&w)+i);
      d=fabs(d);
      if (d>(*d1))
         *d1=d;
   }

   det1.re=
      ((*u).c22.re*(*u).c33.re-(*u).c22.im*(*u).c33.im)-
      ((*u).c23.re*(*u).c32.re-(*u).c23.im*(*u).c32.im);
   det1.im=
      ((*u).c22.re*(*u).c33.im+(*u).c22.im*(*u).c33.re)-
      ((*u).c23.re*(*u).c32.im+(*u).c23.im*(*u).c32.re);
   det2.re=
      ((*u).c21.re*(*u).c33.re-(*u).c21.im*(*u).c33.im)-
      ((*u).c23.re*(*u).c31.re-(*u).c23.im*(*u).c31.im);
   det2.im=
      ((*u).c21.re*(*u).c33.im+(*u).c21.im*(*u).c33.re)-
      ((*u).c23.re*(*u).c31.im+(*u).c23.im*(*u).c31.re);    
   det3.re=
      ((*u).c21.re*(*u).c32.re-(*u).c21.im*(*u).c32.im)-
      ((*u).c22.re*(*u).c31.re-(*u).c22.im*(*u).c31.im);
   det3.im=
      ((*u).c21.re*(*u).c32.im+(*u).c21.im*(*u).c32.re)-
      ((*u).c22.re*(*u).c31.im+(*u).c22.im*(*u).c31.re);

   det.re=
      ((*u).c11.re*det1.re-(*u).c11.im*det1.im)-
      ((*u).c12.re*det2.re-(*u).c12.im*det2.im)+
      ((*u).c13.re*det3.re-(*u).c13.im*det3.im);
   det.im=
      ((*u).c11.re*det1.im+(*u).c11.im*det1.re)-
      ((*u).c12.re*det2.im+(*u).c12.im*det2.re)+
      ((*u).c13.re*det3.im+(*u).c13.im*det3.re);

   *d2=0.0f;
   d=fabs(det.re-1.0f);
   if (d>(*d2))
      *d2=d;
   d=fabs(det.im);
   if (d>(*d2))
      *d2=d;
}


int main(int argc,char *argv[])
{
   int my_rank,count;
   float beta,d1,d2,dmax1,dmax2,dmax1_all,dmax2_all;
   su3 *u,*um;
   FILE *flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
   
      printf("\n");
      printf("Check of the update program (SU(3) normalization)\n");
      printf("-------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);      
      fflush(flog);
   }
   
   beta=5.7f;   
   start_ranlux(0,123456);   
   geometry();
   alloc_u();
   random_u();   

   dmax1=0.0f;
   dmax2=0.0f;

   for (count=0;count<RENORMALIZE;count++)
      update(beta,1,2,RENORMALIZE);
   
   for (count=0;count<RENORMALIZE;count++)
   {
      update(beta,1,0,RENORMALIZE+1);

      u=pu[VOLUME/2][0];
      um=u+4*VOLUME;
      
      for (;u<um;u++)
      {
         dev(u,&d1,&d2);

         if (d1>dmax1)
            dmax1=d1;
         if (d2>dmax2)
            dmax2=d2;            
      }
   }

   MPI_Reduce(&dmax1,&dmax1_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Reduce(&dmax2,&dmax2_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("beta = %.2f\n\n",beta);
   
      printf("In the course of %d unrenormalized heatbath sweeps: \n",
             RENORMALIZE); 
      printf("Maximal deviation from unitarity = %.2e\n",dmax1_all);
      printf("Maximal deviation from det{U}=1  = %.2e\n\n",dmax2_all);
   }

   update(beta,0,0,1);

   dmax1=0.0f;
   dmax2=0.0f;
   u=pu[VOLUME/2][0];
   um=u+4*VOLUME;
      
   for (;u<um;u++)
   {
      dev(u,&d1,&d2);

      if (d1>dmax1)
         dmax1=d1;
      if (d2>dmax2)
         dmax2=d2;  
   }   

   MPI_Reduce(&dmax1,&dmax1_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Reduce(&dmax2,&dmax2_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {    
      printf("After renormalization: \n");
      printf("Maximal deviation from unitarity = %.2e\n",dmax1_all);
      printf("Maximal deviation from det{U}=1  = %.2e\n\n",dmax2_all);   
   }   

   dmax1=0.0f;
   dmax2=0.0f;
   
   for (count=0;count<RENORMALIZE;count++)
   {
      update(beta,0,1,RENORMALIZE+1);

      u=pu[VOLUME/2][0];
      um=u+4*VOLUME;
      
      for (;u<um;u++)
      {
         dev(u,&d1,&d2);

         if (d1>dmax1)
            dmax1=d1;
         if (d2>dmax2)
            dmax2=d2;  
      }
   }   

   MPI_Reduce(&dmax1,&dmax1_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Reduce(&dmax2,&dmax2_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {   
      printf("In the course of %d unrenormalized over-relaxation sweeps: \n",
             RENORMALIZE); 
      printf("Maximal deviation from unitarity = %.2e\n",dmax1_all);
      printf("Maximal deviation from det{U}=1  = %.2e\n\n",dmax2_all);
   }
   
   update(beta,0,0,1);

   dmax1=0.0f;
   dmax2=0.0f;
   u=pu[VOLUME/2][0];
   um=u+4*VOLUME;
      
   for (;u<um;u++)
   {
      dev(u,&d1,&d2);

      if (d1>dmax1)
         dmax1=d1;
      if (d2>dmax2)
         dmax2=d2;  
   }   

   MPI_Reduce(&dmax1,&dmax1_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Reduce(&dmax2,&dmax2_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {    
      printf("After renormalization: \n");
      printf("Maximal deviation from unitarity = %.2e\n",dmax1_all);
      printf("Maximal deviation from det{U}=1  = %.2e\n\n",dmax2_all);   
      fclose(flog);
   }   

   MPI_Finalize();    
   exit(0);
}
