
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Statistical test of the cabmar program [field equation]
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "update.h"
#include "global.h"


static void random_staples(su3 *v)
{
   int i;
   float *r;

   r=(float*)(v);
   gauss(r,18);

   for (i=0;i<18;i++)
      r[i]*=3.0f;
}


static su3 random_lie(void)
{
   int i;
   float *x1,*x2,tr;
   su3 x,y;

   random_su3(&x);
   _su3_dagger(y,x);
   
   for (i=0;i<18;i++)
   {
      x1=(float*)(&x)+i;
      x2=(float*)(&y)+i;
      (*x1)-=(*x2);
   }

   tr=x.c11.im+x.c22.im+x.c33.im;
   tr/=3.0f;

   x.c11.im-=tr;
   x.c22.im-=tr;
   x.c22.im-=tr;

   return(x);
}
   
   
int main(int argc,char *argv[])
{
   int my_rank,n,nmax;
   int i,seed;
   float beta;
   double *a,abar,sigma,abar_all,sigma_all;
   su3 u,v,x,vdag,xu,xuv;
   FILE *flog=NULL,*fin=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);
      fin=freopen("check4.in","r",stdin);
      
      printf("\n");
      printf("Statistical test of the program `cabmar'\n");
      printf("----------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);           

      scanf("beta %f nmax %d seed %d\n",&beta,&nmax,&seed);
      fclose(fin);

      printf("beta = %.1f\n",beta);
      printf("nmax = %d\n",nmax);
      printf("seed = %d\n",seed);
      fflush(flog);
   }

   MPI_Bcast(&beta,1,MPI_FLOAT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmax,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&seed,1,MPI_INT,0,MPI_COMM_WORLD);
   
   a=(double*)(malloc(nmax*sizeof(double)));
   start_ranlux(0,seed);

   random_su3(&u);
   random_staples(&v);
   x=random_lie();
   _su3_dagger(vdag,v);

   for (i=0;i<100;i++)
      cabmar(beta,&u,&v,0);

   for (n=0;n<nmax;n++)
   {
      for (i=0;i<10;i++)
      {
         cabmar(beta,&u,&v,0);
         cabmar(beta,&u,&v,1);
         cabmar(beta,&u,&v,1);
      }

      _su3_times_su3(xu,x,u);
      _su3_times_su3(xuv,xu,vdag);

      a[n]=(double)(3.0f*(xu.c11.re+xu.c22.re+xu.c33.re)+
                    beta*(u.c11.re+u.c22.re+u.c33.re)*
                    (xuv.c11.re+xuv.c22.re+xuv.c33.re));
   }

   abar=0.0;
   sigma=0.0;
   
   for (n=0;n<nmax;n++)
   {
      abar+=a[n];
      sigma+=a[n]*a[n];
   }

   MPI_Reduce(&abar,&abar_all,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Reduce(&sigma,&sigma_all,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   
   abar=abar_all/(double)(nmax*NPROC);
   sigma=sigma_all/(double)(nmax*NPROC);
   sigma=sqrt(sigma/(double)(nmax*NPROC));
 
   if (my_rank==0)
   {    
      printf("\n");
      printf("|<Equation>| = %.3e (should be zero within errors)\n",fabs(abar));
      printf("sigma        = %.3e \n\n",sigma);   
      fclose(flog);
   }   

   MPI_Finalize();
   exit(0);
}
